/*
	PC89LINK: HDOS <-> MSDOS File Transfer Package

	Copyright (c) 1989 Lindley Systems
			   4257 Berwick Place
			   Woodbridge, VA 22192


*/

/* Program control */

	/* Version number is Major * 100 + Minor */
	#define VERSION 201
	/* If TRUE, enable debugging information: */
	#define DEBUG 0
	/* New information packet contains additional information */
	#define EXTRA_INFO 0

/* Include Files */

	#include <stdio.h>

	#include <ctype.h>
	/* Don't want macro versions of these: */
	#undef toupper
	#undef tolower

	#include <stdlib.h>
	#include <dos.h>
	#include <conio.h>
	#include <direct.h>
	#include <signal.h>
	#include <string.h>
	#include "screen.h"

/* External variables & functions */

	extern int showfile (char *filter, unsigned attributes);

	/* Serial I/O Support -- ZPORTS.C or PCPORTS.C */

	extern void initserial();	/* Set up Interrupt masks etc */
	extern void resetserial();	/* Remove interrupt masks */

	extern void setup();		/* Initialize serial port */
	extern void loopback(int type);	/* Set/Clear local loopback */
	extern int nports();		/* How many serial ports? */
	extern void nput (char c);	/* Put character */
	extern int tget();		/* Get character, or timeout if none */
	extern unsigned char nget();	/* Wait for character */
	extern int tstport();		/* Did we get a character? */
	extern int maxbaud;		/* how many different rates? */
	extern int baudent;		/* default baud index */
	extern long baudtable[0];	/* Baud rate table */
	extern int rcv_pkt(int pktlen, char *buf);
				/* Receive whole packet */
/* Prototypes: */

	unsigned char nget ();
	void wait (int msec);
	int sendpkt(char *packet, int pktlen, int pkttyp);
	int getpkt(char	*packet,int *pktlen, int lookfor);
	void init_config();
	void save_config();
	void test_port();

/* Definitions */

	#define MENU_TOP 5
	#define MENU_HEIGHT 9

	#define STAT_TOP 16
	#define STAT_HEIGHT 5

	#define	BLKSIZ 128

	/* Packet Types:

			PKTF	Packet contains header (file name)
			PKTI	Packet is integer (header, File Information)
			PKTD	Packet contains file data
			PKTEOF	Packet indicates end of file

	   Receive-Packet errors:
			PKTERR	Packet contains errors
			PKTNO	Packet not successfully received
			PKTABT	User aborted

	   Send-Packet return values:
			PKTOK	Packet was transferred successfully
			PKTNO	Packet not successfully transferred
			PKTCAN	Receiver requests file restart
			PKTABT	User aborted

	*/

	#define PKTF	0
	#define PKTI	1
	#define PKTD	2
	#define PKTEOF	3
	#define PKTERR	4
	#define PKTNO	5
	#define PKTOK	6
	#define PKTCAN	7
	#define PKTABT	8
	#define PKTMAX	PKTABT

	/* The SYN character is sent from the receiver to the sender
	   to indicate that it is ready to receive a packet */

	#define SYN 22


	/* Packet Response: sent from receiver to sender

			ACK	Packet received correctly
			NAK	Re-send packet; errors encountered
			CAN	Abort operation
	*/

	#define ACK 6
	#define NAK 15
	#define CAN 24

	/* Wait	1/10 second per	packet,	1/20 sec. per character	*/
	#define	PKTWAIT	100
	#define	TIMEOUT	20
	#define SENDWAIT 50

/* Structures */

	#if EXTRA_INFO
	  struct xfile_info
	    {	int link_version;	/* Version of PC89LINK */
		int bin_flag;		/* Binary file flag */
		int file_blocks;	/* # of 256-byte blocks in file */
		int file_date;
		int file_time;
		int file_attrs;
		int spares[10];
	    } cur_file_info;
	  #define PKTI_SIZE sizeof (struct xfile_info)
	#else
	  #define PKTI_SIZE 2
	#endif

	#define DISKBUF_SIZE 10240

	#define MODE_TEXT 0
	#define MODE_BIN  1
	#define MODE_AUTO 2

/* Global variables */

	char *config_file = "PC89LINK.CFG";

	int packet_number;	/* Packet number, to ensure synchronization
				   of file xfer */

	int binary = 0;		/* Binary file?	*/
	int blocks;		/* Number of 256-byte blocks in file */
	int file_time,file_date;

	char buf[256];		/* receiving buffer */
	char diskbuf[DISKBUF_SIZE];	/* disk buffer */
	int diskbuf_pos = 0;

	int portnum = 0;	/* 'portnum' is n of COMn: */

	int binary_config = MODE_AUTO;	/* Assume automatic */

	char rcvdev[65]="";
	int last_status;	/* Used by PKTSTAT */
	int menu_up=0;

	/* Statistics */

	unsigned long send_bytes=0;	/* Number of transmitted bytes */
	unsigned long send_hseconds=0;	/* Elapsed time of transmission */
	unsigned long send_packets=0;	/* Total number of packets */
	unsigned long send_errors=0;	/* Number of retransmitted packets */

	unsigned long rcv_bytes=0;
	unsigned long rcv_hseconds=0;
	unsigned long rcv_packets=0;	/* Total number of packets */
	unsigned long rcv_errors=0;	/* Number of retransmissions we've received */

	unsigned long start_time=-1L;	/* Starting time of current operation */
					/* (Initally illegal) */

	int int_errcode;		/* Last critical error */

/* Functions: */

void critical_handler(unsigned deverror, unsigned errcode, unsigned far *devhdr)
{
	int_errcode = errcode & 15;
	_hardretn(int_errcode+0x13);
}

/*	give_error

	Display the given message.
*/

void give_error(char *text)
{  static char stars[]=" * * * ";

	*strchr(text,'\n') = '\0';	/* Remove newlines */

	wstr(stars);
	wstr(text);
	wstr(stars);
}

/*	give_sys_error

	Displays the given error, plus the last system library error.
*/

void give_sys_error(char *text)
{
	if (errno >= 0)
	  give_error(_strerror(text));
	else
	  switch (int_errcode)
	    { case 0:
		give_error("Write Protected");
		break;
	      case 2:
		give_error("Drive Not Ready");
		break;
	      default:
		give_error("General Error");
		break;
	    }
}

int okay()
{	printf("? (Y/N) ");
	while (1)
	  switch(toupper(gtkey()))
	    { case 'Y':
		return(1);
	      case 'N':
		return(0);
	    }
}

erasemenu()
{
	erase(MENU_TOP,20,MENU_TOP+MENU_HEIGHT,60);
	erase(STAT_TOP,20,STAT_TOP+STAT_HEIGHT,60);
	menu_up=0;
	setcur(5,0);
}

drawtop()
{ char *p;
	cls();
	box(0,0,3,79,1);
	setcur(0,29);
	wstr("\1PC89-LINK Version ");
	wchar(VERSION/100+'0');
	wchar('.');
	wchar((VERSION/10)%10+'0');
	wchar(VERSION%10+'0');
	wchar(2);
	setcur(1,1);
	/* If rcvdev is empty, determine current directory.
	   If error, set rcvdev empty again. */
	if ((*rcvdev) == '\0')
	  { if (getcwd(rcvdev,65)==NULL)
	      *rcvdev='\0';			/* If error, no directory */
	    for (p=rcvdev; *p != '\0'; p++)
	      { if (*p < '#' || *p > '\\')	/* Illegal name? */
	          { rcvdev[3] = '\0';		/* Leave only "X:\" */
		    break;
		  }
	      }
	  }
	printf("Port: COM%1d; Speed:%6ld; Directory [%s]",
		portnum+1,baudtable[baudent],rcvdev);
}

drawmenu()
{ long seconds;

  if (!menu_up)
      {	menu_up=1;
        drawtop();
	box(MENU_TOP,20,MENU_TOP+MENU_HEIGHT,60,2);
	wstr("Press:");
	setcur(7,25);
	wstr("S -- Send File(s)");
	setcur(8,25);
	wstr("R -- Receive File(s)");
	setcur(9,25);
	wstr("D -- Directory");
	setcur(10,25);
	wstr("G -- Go To Subdirectory");
	setcur(11,25);
	wstr("C -- Configure");
	setcur(12,25);
	wstr("T -- Test Connection");
	setcur(13,25);
	wstr("Q -- Quit");

	box (STAT_TOP,20,STAT_TOP+STAT_HEIGHT,60,1);
	setcur(STAT_TOP+1,41);
	wstr("\1  Sent  \2 \1Received\2");
	setcur(STAT_TOP+2,22);
	wstr("Bytes Transferred");
	setcur(STAT_TOP+2,41);
	printf("%8ld",send_bytes);
	printf("%9ld",rcv_bytes);
	setcur(STAT_TOP+3,22);
	wstr("Retransmissions");
	setcur(STAT_TOP+3,41);
	if (send_packets)
	  printf("%7ld%%",send_errors*100L/send_packets);
	setcur(STAT_TOP+3,50);
	if (rcv_packets)
	  printf("%7ld%%",rcv_errors*100L/rcv_packets);
	setcur(STAT_TOP+4,22);
	wstr("Average Bytes/Sec");
	setcur(STAT_TOP+4,41);
	seconds = send_hseconds / 100L;
	if (seconds)
	  printf("%8ld",send_bytes/seconds);
	setcur(STAT_TOP+4,50);
	seconds = rcv_hseconds / 100L;
	if (seconds)
	  printf("%8ld",rcv_bytes/seconds);
      }
}

static void pause()
{	setcur(23,0);
	wstr("\1Press any key to continue:\2");
	gtkey();
}

int menu()
{
  char c;
  int quit, temp;

    quit=0;
    while (!quit)
      {	drawmenu();
	/* Flush type-ahead: */
	while (kbhit())
	  gtkey();
	setcur(6,29);
	c = toupper(gtkey());
	switch(c)
	  { case 'S':
		erasemenu();
		setcur(2,3);
		wstr("\1Sending Files\2 Mode: ");
		switch (binary_config)
		  { case MODE_AUTO:	/* Automatic */
			printf("Automatic Binary/Text");
			break;
		    default:
			printf("Forced to %s",
				binary_config==MODE_TEXT?"TEXT":"BINARY");
		  }
		box(MENU_TOP,10,MENU_TOP+3,70,2);
		wstr("Enter name of file(s) to SEND:");
		setcur(7,12);
		getfname(buf,57);
		erase(5,10,8,70);
		if (*buf)
		  sendfile(buf);
		break;
	    case 'R':
		erasemenu();
		setcur(2,3);
		wstr("\1Receiving Files\2");
		getfile();
		break;
	    case 'G':
		erasemenu();
		showfile("*",_A_SUBDIR);
		box(MENU_TOP,40,MENU_TOP+3,70,2);
		wstr("Enter Subdirectory name:");
		setcur(7,42);
		getfname(buf,27);
		strupr(buf);
		if (*(buf+1) == ':')	/* Gave drive specifier */
		  _dos_setdrive((*buf)&15,&temp);
		if (*buf)
		  { chdir(buf);
		    *rcvdev = '\0';	/* Flag to remind us to get dir. again */
		  }
		break;
	    case 'C':
		erasemenu();
		pickport();
		pickbaud();
		picktype();
		setup();	/* Set up new port */
		save_config();
		break;
	    case 'D':
		erasemenu();
		box(MENU_TOP,10,MENU_TOP+3,70,2);
		wstr("Enter Directory wildcard:");
		setcur(7,12);
		temp=getfname(buf,57);
		erase(5,10,8,70);
		if (temp)
		  { if (!*buf)
			strcpy(buf,"*.*");
		    showfile(buf,_A_NORMAL);
		    pause();
		  }
		break;
	    case 'T':
		erasemenu();
		test_port();
		break;
	    case 'Q':
		erasemenu();
		setcur(5,1);
		quit=1;
	  }
      }
}

/*
	OpenFile

	Attempts to open a file -- looks in the current directory
	first; then looks through all the directories in the Path.
*/

FILE *OpenFile(char *wholename, char *fname)
{ FILE *f;

    if ((f = fopen(fname,"r")) != NULL)
	{ strcpy(wholename,fname);
	  return(f);
	}

    _searchenv(fname,"PATH",wholename);
    if (*wholename)	/* If file found */
	return(fopen(wholename,"r"));
    else
	return(NULL);
}

void init_config()
{ FILE *cfg_file;
  char cfg_name[64];

	if ((cfg_file = OpenFile(cfg_name,config_file)) != NULL)
	  { fscanf(cfg_file,"%d",&portnum);
	    fscanf(cfg_file,"%d",&baudent);
	    fscanf(cfg_file,"%d",&binary_config);
	    fclose(cfg_file);
	  }
}

void save_config()
{ FILE *cfg_file;
  char answer;

	erasemenu();
	drawtop();
	setcur(4,1);
	printf("Save configuration in file %s",rcvdev);
	if (rcvdev[strlen(rcvdev)-1] != '\\')
	  printf("\\");
	printf("%s",config_file);
	if (okay())
	  { setcur(10,1);
	    if ((cfg_file = fopen(config_file,"w")) != NULL)
	      { fprintf(cfg_file,"%d\r\n",portnum);
	        fprintf(cfg_file,"%d\r\n",baudent);
	        fprintf(cfg_file,"%d\r\n",binary_config);
	        fclose(cfg_file);
		printf("Configuration file written.");
	      }
	    else
	      give_sys_error("Can't open file");
	    pause();
	  }
}

void test_stats(long value, int xpos)
{  static char buf[20];

  sprintf(buf,"%6ld",value);
  setcur(15,xpos);
  wstr(buf);
}

void test_port()
{ int index, connected=0, rcv_ch;
  long rcv_count=0, rcv_run=0;
  long err_count=0, missed_count=0;
  int rcv_index;
  int update=0;
  char *string = "ABCDEFGHIJKLMNOPQRSTUVWXYZ\r\nZYXWVUTSRQPONMLKJIHGFEDCBA\r\n";

	setcur(5,0);
	wstr("Testing Port:\r\n\n");

	index=5;
	wstr("Local Loopback Test...");
	while (((rcv_ch = tget()) > 0) && --index)
	  ;
	loopback(1);
	for (index=0; string[index] != 0; index++) {
	  nput(string[index]);
	  if ((rcv_ch = tget()) != string[index]) {
	    printf("Failed! [Received: %d]\n",rcv_ch);
	    printf("Try this test at least twice.  If it continues to fail,\n");
	    printf("it may indicate a problem with the way this computer\n");
	    printf("is configured.  Check memory-resident software such as\n");
	    printf("mouse drivers.\n");
	    pause();
	    return;
	  }
	}
	loopback(0);
	wstr("Passed.");
	setcur(10,0);
	wstr("Remote System Connect Test...");
	setcur(12,34);
	wstr("---Characters---");
	setcur(13,11);
	wstr("Waiting  Connected   Xfer'd Missed Errors");
	setcur(14,11);
	wstr("-------  ---------   ------ ------ ------");
	setcur(15,14);
	wchar('*');
	center(20,"Press ESC key to abort");

	/* Display updates happen at least every 15 errors, 15 misses,
	 * 15 received characters, and at every time-out.
	 */
	index=0;
	while (!kbhit() || (gtkey() != 27)) {
	  if (tstport() < 0) {		/* error */
	    err_count++;
	    if ((err_count & 15) == 0)
	      update++;
	  } else {
	    if ((rcv_ch = tget()) < 0) {	/* timed out */
	      if (connected) {
		missed_count++;
		connected=0;
		setcur(15,14);
		wchar('*');
		setcur(15,24);
		wchar(' ');
		update++;
	      } else {
	        while ((string[index] != 0) && (tstport() <= 0))
	          nput(string[index++]);
	        index=0;
	      }
	    } else {			/* got a character */
	      if (connected == 0) {
		setcur(15,14);
		wchar(' ');
		setcur(15,24);
		wchar('*');
		rcv_index = -1;		/* haven't sync'ed up */
		rcv_run = 0;		/* restart run */
		connected++;
		update++;
	      }
	      if (rcv_index <0) {
	        if (rcv_ch == string[0])
	          rcv_index = 1;	/* sync'ed up! */
					/* (next char will be 2d in string) */
	      } else {
		if (rcv_ch == string[rcv_index++]) {
		  rcv_run++;		/* this run */
		  if ((rcv_count++ & 31) == 0)
		    update++;
		} else {
		  rcv_index = -1;	/* resynch */
		  rcv_run = 0;		/* restart run */
		  if (!iscntrl(rcv_ch))
		    if (isalpha(rcv_ch) || isspace(rcv_ch)) {
		      missed_count++;	/* missed a character */
		      if ((missed_count & 15) == 0)
			update++;
		    } else {
		      err_count++;	/* bad character */
		      if ((err_count & 15) == 0)
		        update++;
		    }
		}
	      }
	      if (string[rcv_index] == 0)
	        rcv_index = 0;
	    }
	  }
	  if (update) {
	    test_stats(rcv_count,32);
	    test_stats(missed_count,39);
	    test_stats(err_count,46);
	    update=0;
	  }
	  if (connected) {
	    if (string[index] == 0)
	      index=0;
	    nput(string[index++]);
	  }
	  if (rcv_run == 2000) {
	    setcur(10,30);
	    wstr("Passed");
	  }
	}
}

unsigned long tclock()
{ struct dostime_t cur_time;
  unsigned long t;

	_dos_gettime(&cur_time);
	t = (long)cur_time.hour;
	t *= 60L;
	t += (long)cur_time.minute;
	t *= 60L;
	t += (long)cur_time.second;
	t *= 100L;
	t += (long)cur_time.hsecond;
	return(t);
}

void start_xmit()
{
	start_time = tclock();
}

unsigned long stop_xmit()
{ unsigned long stop_time;

	stop_time = tclock() - start_time;	/* Elapsed seconds */
	if (stop_time < 0L)			/* Rolled over midnight? */
	  stop_time += 8640000;			/* Compensate */

	/* If start_time was illegal, assume no elapsed time */
	return (start_time == -1L?0L:stop_time);
}

/*
	'which' parameter determines action:

	 2	Show Byte count
	 1	Show File Size & Mode
	 0	Show filename
	-1	Initialize screen
	-2	Show 'Connecting...'
	-3	Show 'Complete'
	-4	Show 'ERROR'
*/

#define WINTOP 5
#define WINLEFT 5
#define WINBOT 19
#define WINRIGHT 75

filestats(char *fname,long bytes,int pctcomp,int binary,int which)
{ int donefile=0;

  if (which==-1)
    { box(WINTOP-1,WINLEFT-1,WINBOT+1,WINRIGHT+1,1);
      wchar(1);
      erase(WINBOT,WINLEFT,WINBOT,WINRIGHT);
      setcur(WINBOT,WINLEFT);
      wstr("Filename           Size Mode   Xfer Retry Wait  Bytes   %  Status");
/*          01234567890123456789012345678901234567890123456789012345678901234 */
/*                    1         2         3         4         5          6 */
      wchar(2);
    }
  if (fname)
    { /* Print new filename */
      setcur(WINBOT-1,WINLEFT);
      wstr(fname);
    }
  if (which<-1 || which ==1)	/* If which=1, just clears status column */
    { erase(WINBOT-1,WINLEFT+59,WINBOT-1,WINRIGHT);
      setcur(WINBOT-1,WINLEFT+59);
      switch (which)
        { case -2:
	    wstr("Connecting");
	    break;
	  case -3:
	    if (pctcomp > 0)
	      {	wstr("Complete");
		which = 3;
		donefile++;
		break;
	      }
	  case -4:
	    wstr("ERROR");
	    donefile++;
	}
    }
  if (which==1)
    { setcur(WINBOT-1,WINLEFT+17);
      if (bytes<1024)
        bytes=1024;
      printf("%5ldK",bytes/1024L);
      setcur(WINBOT-1,WINLEFT+24);
      wstr(binary?"Binary":"Text");
    }
  if (which>1)
    { setcur(WINBOT-1,WINLEFT+46);
      printf("%7ld %3d",bytes,pctcomp);
    }
  else
    last_status = 999;		/* Display packet status again */

  if (donefile)			/* File complete - leave blank line @ bottom */
    scroll(WINTOP,WINLEFT,WINBOT-1,WINRIGHT,1);
}

/*
	Shows current packet status.
	Argument is:

	 1	Packet rcvd/sent OK
	 0	No packet
	-1	Packet retry

*/

pktstat(int pkt_ok)
{ 
    if (pkt_ok != last_status)
      {	setcur(WINBOT-1,WINLEFT+33);
	wchar(pkt_ok >0?'*':32);
	setcur(WINBOT-1,WINLEFT+38);
	wchar(pkt_ok <0?'*':32);
	setcur(WINBOT-1,WINLEFT+44);
	wchar(pkt_ok==0?'*':32);
	setcur(WINBOT-1,WINLEFT+59);
	last_status = pkt_ok;
      }
}

int getfname(p,limit)
  char *p;
  int limit;
{ char *q, ch;
  int count=0;

	q = p;
	ch=0;
	while (ch != 10)
	  { ch = gtkey();
	    if (ch >= ' ')
	      {	if (count < limit)
	          { putchar(ch);
		    *q++ = ch;
		    count++;
		  }
	      }
	    else
	      switch (ch)
		{ case 10:
		  case 13:
			*q = '\0';
			return(1);
		  case 27:
			*p = '\0';
			return(0);
		  case 8:
		  case 127:
			if (q>p)
			  { q--;
			    count--;
			    printf("\b\040\b");
			  }

		}
	  }
}

/* HDOS and MSDOS date words are basically the same:

	MSDOS: yyyyyyymmmmddddd		year based on 1980
	HDOS:  0yyyyyymmmmddddd		year based on 1970

   The dates and times passed to, and returned by, get_ and set_file_date
   will be in the HDOS format.  This makes it easier on the HDOS system.

   CP/M does not support dates.

*/

int hdos_to_msdos_date(date_word)
  int date_word;
{ int year;

	year = (date_word >> 9);
	year = (year < 10)?0:year-10;	/* Base it on 1980 (if before, =1980)*/
	return ((date_word & 0x1FF) | (year << 9));
}

int msdos_to_hdos_date(date_word)
  int date_word;
{ int year;

	year = (date_word >> 9) + 10;
	return ((date_word & 0x1FF) | (year << 9));
	/* Assume we can use top bit of date, even though HDOS doesn't */
}

/* Get and Set file dates: These functions return 0 if no error. */

int get_file_date(fchan, dos_date, dos_time)
  FILE *fchan;
  int *dos_date, *dos_time;
{ int r;
	/* "fchan-_iob" is a kluge to get file handle number */
	/* NOTE!! uses undocumented _iob array */

	if (r=_dos_getftime(fchan-_iob,dos_date,dos_time)!=0)
		return (r);
#if DEBUG
	printf("\n\nDate=%x Time=%x",*dos_date,*dos_time);
#endif
	*dos_date = msdos_to_hdos_date(*dos_date);
	return(0);
}

int set_file_date(fchan,new_date,new_time)
  FILE *fchan;
{
#if DEBUG
	printf("\n\n\nDate=%x Time=%x",hdos_to_msdos_date(new_date),new_time);
#endif
  if (new_date)	/* Only if valid (non-zero) date */
    return(_dos_setftime(fchan-_iob,hdos_to_msdos_date(new_date),new_time));
}

/* File-Level Communications Routines */

int sendfile(fname)
  char *fname;
{
  int pkl,err,pktnum;
  long bytes;
  char c, *p;
  char mode[8];
  struct find_t a_file;
  FILE *fchan;
  char path[64], *path_end;

	filestats(NULL,-1L,-1,-1,-1);		/* Initialize */

	strcpy(path,fname);
	for (path_end=path+strlen(path)-1;path_end>path;path_end--)
	  if (*path_end == '\\' || *path_end == ':')
	    { path_end++; break; }

	*path_end = '\0';

	if (_dos_findfirst(fname,_A_NORMAL,&a_file))
	  { setcur(WINBOT-2,WINLEFT);
	    give_sys_error(fname);
	    setcur(WINBOT-1,WINLEFT);
	    wstr(" ... press any key:");
	    gtkey();
	    return(0);
	  }
	binary = binary_config;
	switch (binary)
	  { case MODE_TEXT:		/* Force TEXT */
		strcpy(mode,"r");
		break;
	    default:	/* Force BINARY, or AUTO */
		strcpy(mode,"rb");
		break;
	  }
	do
	  {	*path_end = '\0';

#if 0
	printf("\nPath [%s] File [%s]\n",path,a_file.name);
#endif

	        strcat(path,a_file.name);
		fchan = fopen(path,mode);

		/* fchan = fopen(a_file.name,mode); */
		if (fchan == NULL)
			return;

		/* Automatic Text/Binary file determination */
		if (binary == MODE_AUTO)
		  { pkl = fread(buf,1,BLKSIZ,fchan);
		    for (binary=0, p=buf, pktnum=0;
			!binary && pktnum<BLKSIZ; pktnum++)
		      binary |= *p++ & 128;
		    if (!binary)	/* Text file after all */
		      { fclose(fchan);
		        strcpy(mode,"r");
			/* Re-open in text mode */
			fchan = fopen(a_file.name,mode);
		      }
		    else
		      binary = MODE_BIN;
		  }

		filestats(a_file.name,-1L,-1,-1,-2);
		pktnum = 0;

		if ((err=sendpkt(a_file.name,strlen(a_file.name)+1,PKTF))!=PKTOK)
		  goto senderr;

		/* Determine length of file in 256-byte blocks */
		fseek(fchan,0L,SEEK_END);
		bytes = ftell(fchan);
		blocks = (int)(bytes >> 8);

		get_file_date(fchan, &file_date, &file_time);

		filestats(NULL,-1L,-1,-1,0);
		fseek(fchan,0L,SEEK_SET);	/* Move to start of file again */
		filestats(NULL,bytes,-1,binary,1);

		bytes = 0;

		start_xmit();

#if EXTRA_INFO
		/* Set up file information record */

		cur_file_info.link_version = VERSION;
		cur_file_info.bin_flag     = binary;
		cur_file_info.file_blocks  = blocks;
		cur_file_info.file_date	   = file_date;
		cur_file_info.file_time	   = file_time;
		cur_file_info.file_attrs   = 0;

		if ((err=sendpkt((char *)&cur_file_info,
			PKTI_SIZE,PKTI)) != PKTOK)
		    goto senderr;
#else
		if ((err=sendpkt((char *)&binary,PKTI_SIZE,PKTI)) != PKTOK)
		    goto senderr;
		if ((err=sendpkt((char *)&blocks,PKTI_SIZE,PKTI)) != PKTOK)
		    goto senderr;

		if ((err=sendpkt((char *)&file_time,PKTI_SIZE,PKTI)) != PKTOK)
		    goto senderr;
		if ((err=sendpkt((char *)&file_date,PKTI_SIZE,PKTI)) != PKTOK)
		    goto senderr;
#endif

		while (pkl=fread(buf,1,BLKSIZ,fchan))
		  if ((err=sendpkt(buf,pkl,PKTD))!=PKTOK)
		      goto senderr;
		  else
		    if (kbhit())
		      { sendpkt(buf,0,PKTABT);
		        err = PKTABT;
		        goto senderr;
		      }
		  else
		    { bytes += (long)pkl;
		      if ((++pktnum % 8)==0)
		        filestats(NULL,bytes,
			  (int)(((long)bytes*100)/((long)(blocks+1)*256)),-1,2);
		    }
	
		err = sendpkt(buf,0,PKTEOF);
	
      senderr:
		send_hseconds += stop_xmit();

		if (pktnum)		/* Show OK/ERROR if have xmitted part */
		  filestats(NULL,bytes,(err==PKTOK)?100:-1,-1,-3);
		fclose(fchan);
		if (err == PKTABT)	/* PKTABT if user aborted procedure */
		  return;
		if (err == PKTOK)	/* PKTOK means entire file xfer is OK */
		  { /* Get next file; if none, break out */
		    if (_dos_findnext(&a_file))
		      break;
		  }
		else
		  { /* If err != PKTOK we'll re-send this file. */
		    sendpkt(buf,0,PKTABT);
		  }
	  }
	while (!kbhit());
	if (err)
	  { setcur(2,3);
	    wstr("\1Error sending file ");
	    wstr(a_file.name);
	    wchar(2);
	  }
}

int bfwrite(int size, FILE *stream)
{
	if (size == 0)
	  { diskbuf_pos = 0;			/* Start new file */
	    return(1);
	  }
	if (diskbuf_pos + size >= DISKBUF_SIZE || /* Would overflow our buffer */
		size <0 )			  /* or Forced buffer flush */
	  { fwrite(diskbuf,diskbuf_pos,1,stream);
	    diskbuf_pos = 0;
	  }
	if (size > 0)
	  { memcpy(&diskbuf[diskbuf_pos],buf,size);
	    diskbuf_pos += size;
	  }
}

getfile()
{
  int pkl,pktnum,err,file_opened;
  long bytes,fsize;
  FILE *fchan;
  char *p,*fname,fnamebuf[24];


    filestats(NULL,-1L,-1,-1,-1);

    cursize(0);			/* Turn off cursor */
    while (1)
      {	box(WINBOT+2,20,WINBOT+5,60,2);
	center(WINBOT+3,"Waiting for File...");
	center(WINBOT+4,"(Press any key to abort)");
	file_opened = 0;

	while (!kbhit() && getpkt(buf,&pkl,PKTF)!=PKTF)
	  ;

	if (kbhit())
	  break;

	p = fname = buf;
	while (*p)
	  { if ((*p < 33) || (*p > 91))
		goto rcverr;		/* Illegal filename */
	    if (*p++ == ':')
		fname = p;
	  }

	strcpy(fnamebuf,fname);
	erase(WINBOT+2,20,WINBOT+5,60);
	filestats(fname,-1L,-1,-999,0);

	start_xmit();

#if EXTRA_INFO
	if ((err=get_info_packet(NULL))!=PKTI)
		goto rcverr;

	binary = cur_file_info.bin_flag;
	blocks = cur_file_info.file_blocks;
	file_date = cur_file_info.file_date;	
	file_time = cur_file_info.file_time;

#else
	/* Get binary flag & # of blocks */
	if ((err=get_info_packet(&binary))!=PKTI)
		goto rcverr;

	if ((err=get_info_packet(&blocks))!=PKTI)
		goto rcverr;

	if ((err=get_info_packet(&file_time))!=PKTI)
		goto rcverr;

	if ((err=get_info_packet(&file_date))!=PKTI)
		goto rcverr;
#endif

	if (blocks<1)
		blocks=1;	/* Very small CP/M files return size=-1!!! */
	fsize = (long)(blocks+1)*256;

	fchan = fopen(fname,binary?"wb":"w");
	file_opened = 1;
	bytes = 0;
	pktnum = 0;
	filestats(NULL,fsize,0,binary,1);

	bfwrite(0,fchan);		/* Start new file */
	while ((err=getpkt(buf,&pkl,PKTD))==PKTD) /* While receiving data */
	  { bfwrite(pkl,fchan);
	    bytes+=(long)pkl;
	    if (pkl == -1)
		break;
	    if (!(++pktnum % 8))
	      filestats(NULL,bytes,
		(int)(((long)bytes*100)/fsize),-1,2);
	  }
	
rcverr:
	filestats(NULL,bytes,(err==PKTEOF)?100:-1,-1,-3);

	rcv_hseconds += stop_xmit();

	if (file_opened)
	  { bfwrite(-1,fchan);		/* Flush internal buffers */
	    fclose(fchan);
	    fchan = fopen(fnamebuf,"r");
	    set_file_date(fchan,file_date,file_time);	/* Set file date/time */
	    fclose(fchan);
	  }
	file_opened = 0;
	/* If file is not completely transferred, delete it! */
	if (err == PKTABT || err == PKTCAN)
	  remove(fnamebuf);
      }
    cursize(1);			/* Re-enable cursor */
}

/* Packet-Level	Communications Routines	*/

/*	SENDPKT - Send a packet to the remote node.

	PktTyp should be:
		PKTF	File name
		PKTI	Header information
		PKTD	File data
		PKTEOF	End of file

	Returns:
		PKTOK	Packet sent OK
		PKTNO	No packet was sent from the remote node
		PKTCAN	Improper packet type received: remote
			operation cancelled (restart file)
		PKTABT	User abort
*/
int sendpkt(packet,pktlen,pkttyp)
  char *packet;
  int pktlen, pkttyp;
{
  int i,s;
  int stimes;
  char *p;


  if (pkttyp == PKTF)
    packet_number = 0;		/* File name is packet # 0 */
  else
    { packet_number++;		/* Else go to next packet */
      packet_number &= 127;
    }

  for (stimes =0; stimes < TIMEOUT; stimes++)
    {	do
	  { i=tget();
	  }
	while ((i != SYN) && (i >= 0));
	/* Look for synchronization character.
	   Toss out other characters.
	   Stop if we time out. */
	if (i != SYN)		/* Couldn't synchronize */
	  { pktstat(0);
	    if (kbhit())
	      return(PKTABT);
	    continue;
	  }
	s=0;
	nput(pkttyp);
	while (tstport() > 0)	/* Flush UART's buffer */
	  nget();
	nput(packet_number);
	nput(pktlen);

	for (i=0,p=packet; i<pktlen; i++)
	  {	s += (unsigned)*p;
		nput(*p++);
	  }
	/* If tstport() is already true (even before checksum), it means
	   other end was confused by our header.  Re-send packet. */
	if (tstport() > 0)
	  { send_errors++;
	    continue;
	  }
#ifdef LONGCHECK
	nput(s >> 8);		/* 16-bit checksums */
#endif
	nput(s & 255);

	switch (tget())
	  { case ACK:		/* Right answer */
		send_bytes+=(long)pktlen;
		send_packets++;
		pktstat(1);
		return(PKTOK);
	    case NAK:		/* Wrong answer -- retry immediately */
		pktstat(-1);
		send_errors++;
		break;
	    case CAN:		/* Remote cancel request */
		pktstat(-1);
		return(PKTCAN);
	    default:		/* No response -- resynchronize */
		pktstat(0);
		send_errors++;
		wait(SENDWAIT);
	  }
    }
  return(PKTNO);		/* Sent	it lots	of times, never	succeeded */
}

/* TGET - Timeout Get character. Return character, or -1 if none */

unsigned int get_pkt_type()		/* Get packet type, *SHORT* time-out */
{ int t=500;				/* Countdown delay */
	while (tstport() <= 0)
	  if (--t ==0)
	    return(-1);			/* No character */
	return(nget());			/* Return char */
}

/* Use this routine to get an information packet containing an
   integer. */

int get_info_packet(int *iptr)
{ int ibuf[128];	/* So even if too big things will be OK */
  int err,pkl;

#if EXTRA_INFO
	err = getpkt((char *)ibuf,&pkl,PKTI);
	if (pkl != sizeof (struct xfile_info))
	  return(PKTNO);
	memcpy((char *)&cur_file_info,
		(char *)ibuf,
		sizeof (struct xfile_info));
	return(err);
#else
	err = getpkt((char *)ibuf,&pkl,PKTI);
	*iptr = ibuf[0];
	return(err);
#endif
}

/*	GETPKT - Get a packet from the remote node.

	The 'Lookfor' parameter indicates what type of packet to
	look for:

		PKTF	File name only
		PKTI	Header info packets only
		PKTD	Data or end-of-file packets only

	Returns:
		PKTF	File name
		PKTI	Header information
		PKTD	Received packet was file data
		PKTEOF	Received packet indicated end of file
		PKTERR	Received packet contained errors
		PKTCAN	Unable to receive correct packet from other end.
			Operation cancelled (restart file)
*/

int getpkt(char	*packet,int *pktlen, int lookfor)
{
  int i,s,check,rtimes,packet_type,packet_len,rcv_pkt_num;
  int ignore_pkt,abort;
  char *p;

  if (lookfor == PKTF)
    packet_number = 0;		/* File name is packet # 0 */
  else
    { packet_number++;		/* Else go to next packet */
      packet_number &= 127;	/* Make 0..127 */
    }

  for (rtimes =0; rtimes < TIMEOUT*2; rtimes++)
      {	while (tstport() > 0)			/* Flush buffer */
	  nget();
#if DEBUG
	setcur(20,1);
	printf("Rtimes = %2d",rtimes);
#endif

	nput(SYN);				/* Ready to receive */

	/* Get packet type and length; show 'wait' if either missing */
	if (((packet_type = tget()) <0) ||
	    ((rcv_pkt_num = tget()) <0) ||
	    ((packet_len  = tget()) <0))
	  { pktstat(0);
	    if (kbhit())
	      return(PKTABT);			/* ESC hit -- abort */
	    continue;				/* Timed out */
	  }


/*
	Exceptional packet acceptance if:

		* Looking for Data, but find EOF with 0 length
		* Abort or Cancel packet received

	If so, pretend we were looking for this type of packet!
*/

	if ((lookfor == PKTD && packet_type == PKTEOF && packet_len == 0) ||
	    packet_type == PKTABT || packet_type == PKTCAN)
	  lookfor = packet_type;

/*	If we receive a packet which we have already received,
	acknowledge it and ignore it. (Don't ignore packets passing
	error information.)
*/

	if (rcv_pkt_num > 120 && packet_number <10)
	  rcv_pkt_num -= 128;	/* 127 -> -1 for sequencing */

	ignore_pkt = (rcv_pkt_num<packet_number && 
		      rcv_pkt_num>packet_number-10 &&
		      packet_type != PKTABT && packet_type != PKTCAN);

/*
	Normal Packet checks:
	
	Packet of desired type
		(okay if PKTEOF when looking for PKTD)
		(EOF packets must have 0 length)
	All packets must be less than BLKSIZ long
	Information packets are 2 bytes long

	-- If not, tell other end to retry packet.
*/

	if (!ignore_pkt)
	  if ((packet_type != lookfor) ||
	    (packet_len > BLKSIZ) ||
	    (lookfor == PKTI && packet_len != PKTI_SIZE) ||
	    (rcv_pkt_num != packet_number) )
	  { nput(NAK);
	    while (tget() >=0)		/* throw away packet */
	     {}
	    pktstat(-1);

#if DEBUG
		setcur(21,1);
		printf ("Rcvd Type= %d; Look for %d; Length=%3d *IGNORED*",
			packet_type,lookfor,packet_len);
		clear_eol();
#endif

	    rcv_errors++;		/* Show retransmission */
	    continue;
	  }
	
	/* Receive data part of packet */
	s = rcv_pkt(packet_len, packet);	/* -1 for timeout */
						/* -2 for error */
	switch (s)
	  { case -1:				/* -1 = Timed out */
		rcv_errors++;
		pktstat(0);
		continue;
	    case -2:				/* -2 = I/O error */
		i = 0;				/* Show error */
		break;
	    default:				/* 0  = no error */
	#ifdef LONGCHECK
		check = tget();				/* Top 8 Bits */
		if (check > 0)
		  { check <<= 8;			/* Shift into place */
		    check += tget();			/* Get bottom 8 bits */
		  }
	#else
		check =	tget();				/* Get checksum	*/
	#endif
		for (p=packet, i=packet_len-1, s=0; i>=0; i--)
			s += *p++;

		i = ((s	& 255) == check);

		*pktlen = packet_len;

	  }
#if DEBUG
	setcur(21,1);
	printf ("Rcvd Type= %d; Look for %d; Length=%3d ACK=%3d",
		packet_type,lookfor,packet_len,i);
	clear_eol();
#endif

	if (kbhit())
	  { nput(CAN);		/* cancel operation */
#if DEBUG
	    setcur(21,1); wstr("** CANCELLED **"); clear_eol();
#endif
	    return(PKTCAN);
	  }
	nput(i?ACK:NAK);

	if (i)
	  { rcv_bytes+=packet_len; rcv_packets++; }
	else
	  rcv_errors++;

	pktstat(i?1:-1);

	if (i && !ignore_pkt)
	  return(packet_type);
      }
  return(PKTCAN);
}

#if 0
#define	BDOS_IN	     7

static union  REGS Regs;

/* gtkey ------------------------------------------------------
   Return a key	from the keyboard and map control characters */

int gtkey()
{
    register int c;

    switch(c=bdos(BDOS_IN,0,0) & 0xff)
      {	case 0:		/* Special function key	-- return zero */
	  bdos(BDOS_IN,0,0);
	  return(0);
	case '\r':
	  return('\n');
	case '\n':
	  return('\r');
	default:
	  return(c);
      }
}
#endif

/*
	WAIT - Wait a given # of milliseconds
*/

static int getclk()
{
  struct dostime_t time;

  _dos_gettime(&time);
  return((int)(time.second)*100+time.hsecond);
}

void wait (int msec)
{
  register int waitfor;

  waitfor = getclk()+msec/10;
  if (waitfor >= 6000)	/* Overflowed seconds into minutes */
    waitfor -= 6000;
	/* If we went 'backwards' to the beginning of the next minute,
	   wait	for this minute	to end... */
  while	(getclk() > waitfor)
	;
	/* Now wait until proper moment	in the current second */
  while	(getclk() < waitfor)
	;

}

pickport()
{
	printf("Serial Port: ");
	portnum = tblpick(NULL,nports(),portnum,1);
}

pickbaud()
{
	printf("Transfer Speed: ");
	baudent	= tblpick(baudtable,maxbaud,baudent,0);
}

picktype()
{
  char c;
  int choice=-1;

  printf("When SENDING files:\n\
   (T) Force TEXT mode\n\
   (B) Force BINARY mode\n\
   (A) Automatic selection\n\
\n\
       [Press T, B, or A]: ");
  while (choice < 0)
    {	switch (c = toupper(gtkey()))
	  { case 'T':
		choice=1;
		binary_config = MODE_TEXT;
		break;
	    case 'B':
		choice=1;
		binary_config = MODE_BIN;
		break;
	    case 'A':
		choice=1;
		binary_config = MODE_AUTO;
		break;
	    case 27:
	    case 13:
	    case 10:
		choice=1;
	  }
    }

}

int tblpick(table,max,curent,fmt)
  long *table;
  int max,curent,fmt;
{
  int newnum;
  char c;

	newnum = curent;
	c = 0;
	while (c != 10)
	  { 
	    switch (fmt)
	      { case 1:
		  printf("COM%1d:\b\b\b\b\b",newnum+1);
		  break;
		case 0:
		  printf("%6ld\b\b\b\b\b\b",table[newnum]);
	      }
	    c =	gtkey();
	    switch (c)
	       { case 8:
		  if (--newnum < 0)
		    newnum = max-1;
		  break;
		case 32:
		  if (++newnum >= max)
		    newnum = 0;
		  break;
		case 27:
		  putchar('\n');
		  return(curent);
	      }
	  }
	putchar('\n');
	return(newnum);
}

main ()
{
  signal(SIGINT, SIG_IGN);		/* Disable ^C checking */
  _harderr(critical_handler);

  init_keys();
  init_config();
  initserial();
  setup	();				/* set everything up */
  menu();
  resetserial();
  exit_keys();
}
